/*
 * Copyright 2009 Ilya Epifanov
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *  http://www.apache.org/licenses/LICENSE-2.0
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.xajile.fuchakubutsu.tsuyujimo;

import java.io.IOException;
import java.io.Writer;
import org.joda.time.ReadableDateTime;

public final class FrozenObjectWriter {
  private final Writer out;
  private boolean first = true;
  private char[] encoded = {'_', '0', '0', '0', '0'};

  public FrozenObjectWriter(Writer out) {
    this.out = out;
  }

  private Writer semiColon() throws IOException {
    if (first)
      first = false;
    else
      out.append(';');
    return out;
  }

  public FrozenObjectWriter write(Object v) throws IOException {
    if (v == null) {
      writeNull();
    } else if (v instanceof Integer) {
      writeInt((Integer) v);
    } else if (v instanceof String) {
      writeString((String) v);
    } else if (v instanceof Long) {
      writeLong((Long) v);
    } else if (v instanceof Character) {
      writeChar((Character) v);
    } else if (v instanceof Boolean) {
      writeBool((Boolean) v);
    } else if (v instanceof ReadableDateTime) {
      writeDate((ReadableDateTime) v);
    } else {
      throw new IllegalArgumentException(
              v.getClass().getName() + " isn't supported (yet?)");
    }
    return this;
  }

  public FrozenObjectWriter writeInt(int v) throws IOException {
    semiColon().append("i:");
    if (v >= 0)
      out.append(Integer.toString(v));
    else
      out.append('m').append(Integer.toString(-v));
    return this;
  }

  public FrozenObjectWriter writeLong(long v) throws IOException {
    semiColon().append("l:");
    if (v >= 0)
      out.append(Long.toString(v));
    else
      out.append('m').append(Long.toString(-v));
    return this;
  }

  public FrozenObjectWriter writeString(String v) throws IOException {
    semiColon().append("s:");
    for (int i = 0; i < v.length(); i++)
      writeC(v.charAt(i));
    return this;
  }

  public FrozenObjectWriter writeChar(char v) throws IOException {
    semiColon().append("c:");
    writeC(v);
    return this;
  }

  private void writeC(char c) throws IOException {
    if ((c >= '0' && c <= '9') || (c >= 'a' && c <= 'z') || (c >= 'A' && c <= 'Z') || c == '.') {
      out.append(c);
    } else {
      encoded[1] = hex(c >> 12);
      encoded[2] = hex(c >> 8);
      encoded[3] = hex(c >> 4);
      encoded[4] = hex(c);
      out.write(encoded);
    }
  }

  public FrozenObjectWriter writeBool(boolean v) throws IOException {
    semiColon().append(v ? "b:t" : "b:f");
    return this;
  }

  public FrozenObjectWriter writeNull() throws IOException {
    semiColon().append('n');
    return this;
  }

  public FrozenObjectWriter writeDate(ReadableDateTime v) throws IOException {
    semiColon().append("d:");
    appendI4(v.getYear());
    out.append('_');
    appendI2(v.getMonthOfYear());
    out.append('_');
    appendI2(v.getDayOfMonth());
    out.append("__");
    appendI2(v.getHourOfDay());
    out.append('_');
    appendI2(v.getMinuteOfHour());
    out.append('_');
    appendI2(v.getSecondOfMinute());
    out.append('_');
    appendI3(v.getMillisOfSecond());
    return this;
  }

  private void appendI4(int i4) throws IOException {
    int rem = i4 % 10;
    encoded[4] = (char) (rem + '0');
    i4 = (i4 - rem) / 10;
    rem = i4 % 10;
    encoded[3] = (char) (rem + '0');
    i4 = (i4 - rem) / 10;
    rem = i4 % 10;
    encoded[2] = (char) (rem + '0');
    i4 = (i4 - rem) / 10;
    rem = i4 % 10;
    encoded[1] = (char) (rem + '0');
    out.write(encoded, 1, 4);
  }

  private void appendI3(int i3) throws IOException {
    int rem = i3 % 10;
    encoded[4] = (char) (rem + '0');
    i3 = (i3 - rem) / 10;
    rem = i3 % 10;
    encoded[3] = (char) (rem + '0');
    i3 = (i3 - rem) / 10;
    rem = i3 % 10;
    encoded[2] = (char) (rem + '0');
    out.write(encoded, 2, 3);
  }

  private void appendI2(int i2) throws IOException {
    int rem = i2 % 10;
    encoded[4] = (char) (rem + '0');
    i2 = (i2 - rem) / 10;
    rem = i2 % 10;
    encoded[3] = (char) (rem + '0');
    out.write(encoded, 3, 2);
  }

  private static char hex(int i) {
    i &= 0x0f;
    i += '0';
    if (i > '9')
      i += 39;
    return (char) i;
  }
}
